#include "driver.h"
#include <dc/g2bus.h>
#include <dc/spu.h>
#include <kos.h>

#define SAMPLE_BITS 16
#define SOUND_LEN   0x2000 

//int nominal_sample_rate;
//int soundcard,usestereo;
int attenuation = 0;
int sampleratedetect = 1;

static int stream_cache_stereo;

static int audio_buffer_length;
/* global sample tracking */
static int samples_per_frame;
static UINT32 samples_left_over;
static int samples_this_frame;

/* code from dcgnuboy*/

#define ARM_OFFSET 0xa0800000
#define ARM_OFFSET_B 0x00800000
static volatile uint32 *start  = (uint32*)(0x10000 + ARM_OFFSET);
static volatile uint32 *init_done = (uint32*)(0x10004 + ARM_OFFSET);
static volatile uint32 *play_start = (uint32*)(0x10008 + ARM_OFFSET);
static volatile uint32 *rate = (uint32*)(0x10010 + ARM_OFFSET);
static volatile uint32 *bits = (uint32*)(0x10014 + ARM_OFFSET);
static volatile uint32 *buf_len = (uint32*)(0x10018 + ARM_OFFSET);
static volatile uint32 *position  = (uint32*)(0x10020 + ARM_OFFSET);
#define BASE_ADDRESS 0x11000

static const uint8 stream_driver[] = {
	#include "sound/aica_fw.h"
};

unsigned char use_audio_throttle = 1;
unsigned char use_audio_thread = 0;
unsigned char use_audio_force_mono = 0;

struct PLAY_INFORMATION
{
  INT16 *buffer;
  int channel_buffer;
  unsigned char stop;

  kthread_t *play_thread;
  semaphore_t *play;
  semaphore_t *finished;
} 
PLAY_INFORMATION;

static struct PLAY_INFORMATION *play_information = 0;

/* --------------------------- */
/* Used in audio_thread mode 2 */
/* --------------------------- */

INT16 *play_buffer = 0;
void (*update_audio)(void *buffer) = 0;

/* dc_update_audio_stream ========================================== */
/*                                                                   */
/* ========================================== dc_update_audio_stream */

#define FIFO_WAIT_LEN 128
#define FIFO_WAIT_ADD 512  /* FIFO_WAIT_LEN x 4 */

unsigned char curbuff = 0;

void dc_update_audio_stream(void *buffer)
{
  unsigned long    copy_length = samples_this_frame*(SAMPLE_BITS/8);
  INTERNAL_SIZE_T*      mcsrc1 = (INTERNAL_SIZE_T*) buffer;
  INTERNAL_SIZE_T*      mcsrc2 = (INTERNAL_SIZE_T*) (buffer+copy_length);
  INTERNAL_SIZE_T*      mcdst1 = (INTERNAL_SIZE_T*) (ARM_OFFSET_B+BASE_ADDRESS+(stream_cache_stereo?SOUND_LEN:0)+copy_length*curbuff);
  INTERNAL_SIZE_T*      mcdst2 = (INTERNAL_SIZE_T*) (ARM_OFFSET_B+BASE_ADDRESS+SOUND_LEN+SOUND_LEN+copy_length*curbuff);
  long                     mcn = 0;

  if (use_audio_throttle)
  {
    if (curbuff)
      while(*position>samples_this_frame) {}
    else
      while(*position<samples_this_frame) {};
  }

  copy_length /= sizeof(INTERNAL_SIZE_T);

  if (copy_length >= 8) 
  { 
    mcn = (copy_length-1)/8; 
    copy_length %= 8; 
  }

  if (stream_cache_stereo)
  {
    switch (copy_length) /* Stereo needs left and right */ 
    {
      case 0: for(;;) { *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++; 
      case 7:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++;
      case 6:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++;
      case 5:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++;
      case 4:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++;
      case 3:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++;
      case 2:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++;
      case 1:           *mcdst1++ = *mcsrc1++;
                        *mcdst2++ = *mcsrc2++; if(mcn <= 0) break; --mcn; }
    }
  }
  else
  {
    switch (copy_length) /* Mono needs only one buffer */
    {
      case 0: for(;;) { *mcdst1++ = *mcsrc1++;
      case 7:           *mcdst1++ = *mcsrc1++;
      case 6:           *mcdst1++ = *mcsrc1++;
      case 5:           *mcdst1++ = *mcsrc1++;
      case 4:           *mcdst1++ = *mcsrc1++;
      case 3:           *mcdst1++ = *mcsrc1++;
      case 2:           *mcdst1++ = *mcsrc1++;
      case 1:           *mcdst1++ = *mcsrc1++; if(mcn <= 0) break; --mcn; }
    }
  }
  
  curbuff = (++curbuff)%2;

  return;
}

/* thd_dc_update_audio ============================================= */
/*                                                                   */
/* ============================================= thd_dc_update_audio */

void thd_dc_update_audio(void *v)
{
  while(!(play_information->stop))
  {
    sem_wait(play_information->play);

    if (play_information->stop)
      printf("AUDIO Thread %d stopping\n",play_information->play_thread->tid);
    else if (play_information->buffer)
      dc_update_audio_stream(play_information->buffer);
  }

  sem_signal(play_information->finished);
}

/* dc_signal_audio_thread ========================================== */
/*                                                                   */
/* ========================================== dc_signal_audio_thread */

void dc_signal_audio_thread(void *buffer)
{
  sem_signal(play_information->play);

  return;
}

/* osd_update_audio_stream ========================================= */
/*                                                                   */
/* ========================================= osd_update_audio_stream */

int osd_update_audio_stream(INT16 *buffer)
{
  if (Machine->sample_rate == 0) return 0;

  /* ----------------------------------------------- */
  /* compute how many samples to generate next frame */
  /* ----------------------------------------------- */

	samples_left_over += samples_per_frame;
	samples_this_frame = (UINT32)samples_left_over;
	samples_left_over -= (int)samples_this_frame;

  switch(use_audio_thread)
  {
    case 2:
      if (use_audio_throttle)
      {
        /* ------------------------------------------------- */
        /* Does it look good that we'll be able to copy now? */
        /* ------------------------------------------------- */

        if ((curbuff && *position>samples_this_frame) ||
            (!curbuff && *position<samples_this_frame))
        {
          /* ------------- */
          /* No it doesn't */
          /* ------------- */

          play_buffer = buffer;
          update_audio = &dc_update_audio_stream;
          return samples_this_frame;
        }

        /* ------------------------------ */
        /* Yes, it does - so drop through */
        /* ------------------------------ */
      }
      break;

    case 3:
      play_buffer = buffer;
      update_audio = &dc_update_audio_stream;
      return samples_this_frame;
      break;

    case 4:
      play_information->buffer = buffer;
      update_audio = &dc_signal_audio_thread;
      return samples_this_frame;
      break;
  }

  update_audio = NULL;

  if (!(use_audio_thread & 0x01))
  {
    play_information->buffer = buffer;

    sem_signal(play_information->play);
  }
  else
    dc_update_audio_stream(buffer);

	return samples_this_frame;
}

/* dc_sound_init =================================================== */
/*                                                                   */
/* =================================================== dc_sound_init */

void dc_sound_init (int sample_rate, int sample_bits, int buffer_len)
{
  *init_done = 0; /* need before spu_memload */

  g2_fifo_wait();
  
  spu_disable();
  spu_memload (0, &stream_driver, sizeof(stream_driver));
  spu_enable ();
  
  *rate = sample_rate;
  *bits = sample_bits;
  *buf_len = buffer_len * 2; /* DOUBLE BUFFERING */

  g2_fifo_wait ();
  
  for (;;)
  {
    if (*init_done) break;
    timer_spin_sleep (50);
  }
  
  *start = 1;
  g2_fifo_wait ();
  
  for (;;)
  {
    if (*play_start) break;
    timer_spin_sleep (50);
  }

}

int osd_start_audio_stream(int stereo)
{
	if (Machine->sample_rate == 0) return 0;

	stream_cache_stereo = use_audio_force_mono?0:!(!stereo);

	/* determine the number of samples per frame */
	samples_per_frame = Machine->sample_rate / Machine->drv->frames_per_second;

  //samples_per_frame = (samples_per_frame>>2)<<2;
  samples_per_frame += (samples_per_frame & 0x01);

	/* compute how many samples to generate this frame */
	samples_left_over = samples_per_frame;
	samples_this_frame = (UINT32)samples_left_over;
	samples_left_over -= (int)samples_this_frame;

	audio_buffer_length = samples_this_frame * (SAMPLE_BITS/8);

  if (stream_cache_stereo)
  {
    spu_memset (BASE_ADDRESS+SOUND_LEN, 0, audio_buffer_length);
    spu_memset (BASE_ADDRESS+SOUND_LEN+SOUND_LEN, 0, audio_buffer_length);
  }
  else
    spu_memset (BASE_ADDRESS, 0, audio_buffer_length);

 	dc_sound_init (Machine->sample_rate , SAMPLE_BITS, audio_buffer_length );

  /* -------------------------------------------- */
  /* Launch the sound processing thread if needed */
  /* -------------------------------------------- */

  if (!(use_audio_thread & 0x01))
  {
    if ((play_information = malloc(sizeof(struct PLAY_INFORMATION))))
    {
      memset(play_information,0x00,sizeof(struct PLAY_INFORMATION));

      if ((play_information->play = sem_create(0)) &&
          (play_information->finished = sem_create(0)))
      {
        if ((play_information->play_thread = thd_create(thd_dc_update_audio, NULL)))
        {
          printf("AUDIO thread %d created\n",play_information->play_thread->tid);
        }
        else   /* failed thd_create */
        {
          sem_destroy(play_information->play);
          sem_destroy(play_information->finished);

          free(play_information);

          play_information=0;
        }
      }
      else    /* failed sem_creates */
      {
        free(play_information);

        play_information = 0;
      }
    }

    if (!play_information)
      use_audio_thread = 1;
  }

	return samples_this_frame;
}

void osd_stop_audio_stream(void)
{
	if (Machine->sample_rate == 0) return;

  if (!(use_audio_thread & 0x01))
  {
    play_information->stop = 1;
    sem_signal(play_information->play);
    sem_wait(play_information->finished);

    //thd_destroy(play_information->play_thread);
    sem_destroy(play_information->play);
    sem_destroy(play_information->finished);

    free(play_information);

    play_information = 0;
  }

  update_audio = NULL;

  if (stream_cache_stereo)
  {
    spu_memset (BASE_ADDRESS+SOUND_LEN, 0, audio_buffer_length);
    spu_memset (BASE_ADDRESS+SOUND_LEN+SOUND_LEN, 0, audio_buffer_length);
  }
  else
    spu_memset (BASE_ADDRESS, 0, audio_buffer_length);

  g2_fifo_wait();

	spu_disable();
  
  printf("LEAVING osd_stop_audio_stream\n");
}

/* attenuation in dB */
void osd_set_mastervolume(int _attenuation)
{
	if (_attenuation > 0) _attenuation = 0;
	if (_attenuation < -32) _attenuation = -32;

	attenuation = _attenuation;
}

int osd_get_mastervolume(void)
{
	return attenuation;
}

void osd_sound_enable(int enable_it)
{
}
