#include "driver.h"
#include "osdepend.h"
#include "mamedbg.h"
#include "vidhrdw/vector.h"
#include "ticker.h"
#include "dreamcast.h"
#include <kos.h>
#if defined(WT_ASSERT)
#include <assert.h>
#endif

unsigned long md_calls = 0;

/* input.c */     extern void poll_joysticks(uint32 code);
///* dreamcast.c */ extern int global_colormode;
/* sound.c */     extern INT16 *play_buffer;
/* sound.c */     extern void (*update_audio)(void *buffer);

#define BACKGROUND 0

void (*update_screen)(struct mame_bitmap *bitmap);
//void (*dc_draw_scene)(void);

#if defined(WT_VBLANK_POLL_JOYSTICK)
static int pj_handler = 0;
#endif
static int video_depth,video_fps,video_attributes,video_orientation;
//static int screen_colors;
//static UINT8 *current_palette = 0;
//static unsigned char *dirtycolor = 0;
//static int dirtypalette;

//#ifdef WT_ARCADEMACHINE
//int showfps = 0;
//#else
//int showfps = 1;
//#endif

//int speed = 100;
//TICKER total_ticks_last_cycle = 0;
//TICKER ticks_per_cycle = TICKS_PER_SEC / 60;
int frameskip;
int autoframeskip;
#define FRAMESKIP_LEVELS 12

int vsync_frame_rate;
unsigned char use_video_thread   = 0;
unsigned char use_display_mode   = 0;
//unsigned char interlace        = 0;
unsigned char use_throttle       = 1;                     /* toggled by F10 */
unsigned char use_dma            = 0;
unsigned char screen_orientation = 0;

//static float osd_gamma_correction = 1.0;

static float dc_brightness;
static float brightness;
static float brightness_paused_adjust;

static int frames_displayed;
static int frameskip_counter;
static TICKER start_time;             /* to calculate fps average on exit */
float avg_fps = 0.0;

static UINT16 *current_pens_lookup = 0;
static UINT16 *pens_lookup = 0;
//static UINT16 *debug_pens_lookup = 0;

struct DRAW_INFORMATION
{
  struct mame_bitmap *bitmap;
  unsigned char stop;

  kthread_t *draw_thread;
  semaphore_t *draw;
  semaphore_t *finished;
} 
DRAW_INFORMATION;

static struct DRAW_INFORMATION *draw_information = 0;

//#define SAFETY 16

void osd_reset_counters(void)
{
  frames_displayed = 0;
  start_time = ticker();
}

int32 simple_display2[480];
int32 simple_display[480]; 
int32 simple_bitmap[481];

int32 simple_size = 0;
int32 simple_width = 0;

static pvr_ptr_t dc_texture_addr = 0;
static pvr_poly_hdr_t dc_pvr_poly_header;

float  dc_u1,dc_v1,dc_u2,dc_v2,dc_u3,dc_v3,dc_u4,dc_v4;
int    dc_x1,dc_y1,dc_x2,dc_y2;

/* dc_draw_scene =================================================== */
/*                                                                   */
/* =================================================== dc_draw_scene */

static void dc_draw_scene(void)
{
  pvr_vertex_t vert;

  pvr_scene_begin();

  pvr_list_begin(PVR_LIST_OP_POLY);
                             
  pvr_prim(&dc_pvr_poly_header, sizeof(pvr_poly_hdr_t));

  /* NEED TO HANDLE ROTATION... */

  vert.flags = PVR_CMD_VERTEX;
  vert.x = dc_x1;
  vert.y = dc_y1;
  vert.z = 1;
  vert.u = dc_u1;
  vert.v = dc_v1;
  vert.argb = PVR_PACK_COLOR (1, dc_brightness, dc_brightness, dc_brightness);

  vert.oargb = 0;
  pvr_prim(&vert, sizeof(pvr_vertex_t));

  vert.x = dc_x2;
  vert.u = dc_u2;
  vert.v = dc_v2;
  pvr_prim(&vert,sizeof(pvr_vertex_t));

  vert.x = dc_x1;
  vert.y = dc_y2;
  vert.u = dc_u3;
  vert.v = dc_v3;
  pvr_prim(&vert,sizeof(pvr_vertex_t));

  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.x = dc_x2;
  vert.u = dc_u4;
  vert.v = dc_v4;
  pvr_prim(&vert,sizeof(pvr_vertex_t));

  pvr_list_finish();

  pvr_list_begin(PVR_LIST_TR_POLY);

  /* Actually could overlay text in here... */

  pvr_list_finish();

  pvr_scene_finish();
}

/* update_screen_dma =============================================== */
/*                                                                   */
/* =============================================== update_screen_dma */

static void update_screen_dma(struct mame_bitmap *bitmap)
{
  pvr_txr_load_dma((void *)simple_bitmap[0], dc_texture_addr, simple_size, 1);

  if (use_throttle) pvr_wait_ready();

  return(dc_draw_scene());
}

/* update_screen_non_dma =========================================== */
/*                                                                   */
/* =========================================== update_screen_non_dma */

static void update_screen_non_dma(struct mame_bitmap *bitmap)
{
  int32 *cur_ds2_line = (int32 *)&simple_display2;
  int32 *cur_dsp_line = (int32 *)&simple_display;
  int32 *cur_bmp_line = (int32 *)&simple_bitmap;

  unsigned int n,*s,*d;
	
  if (current_pens_lookup)
	{
    const uint16 *cpl = current_pens_lookup;

    while(*cur_bmp_line)
    {
      n = simple_width;
      s = (unsigned int *)*cur_bmp_line++;
      d = (unsigned int *)*cur_ds2_line++;

      QACR0 = *cur_dsp_line;
      QACR1 = *cur_dsp_line++;

      while(--n)
      {
        d[0] = cpl[s[0]>>16]<<16|cpl[s[0]&0xFFFF];
        d[1] = cpl[s[1]>>16]<<16|cpl[s[1]&0xFFFF];
        d[2] = cpl[s[2]>>16]<<16|cpl[s[2]&0xFFFF];
        d[3] = cpl[s[3]>>16]<<16|cpl[s[3]&0xFFFF];
        d[4] = cpl[s[4]>>16]<<16|cpl[s[4]&0xFFFF];
        d[5] = cpl[s[5]>>16]<<16|cpl[s[5]&0xFFFF];
        d[6] = cpl[s[6]>>16]<<16|cpl[s[6]&0xFFFF];
        d[7] = cpl[s[7]>>16]<<16|cpl[s[7]&0xFFFF];
        asm("pref @%0" : : "r" (d));
        d+=8;
        s+=8;
      }
    }
	}
  else
  {
    while(*cur_bmp_line)
    {
      n = simple_width;
      s = (unsigned int *)*cur_bmp_line++;
      d = (unsigned int *)*cur_ds2_line++;

      QACR0 = *cur_dsp_line;
      QACR1 = *cur_dsp_line++;

      while(--n)
      {
        d[0] = s[0];
        d[1] = s[1];
        d[2] = s[2];
        d[3] = s[3];
        d[4] = s[4];
        d[5] = s[5];
        d[6] = s[6];
        d[7] = s[7];
        asm("pref @%0" : : "r" (d));
        d+=8;
        s+=8;
      }
    }

    d = (unsigned int *)0xe0000000;
    d[0] = d[8] = 0;
  }

  if (use_throttle) pvr_wait_ready();

  return(dc_draw_scene());
}

/* thd_dc_draw_screen ============================================== */
/*                                                                   */
/* ============================================== thd_dc_draw_screen */

static void thd_dc_draw_screen(void *v)
{
  int stopped = 0;

  while(!stopped)
  {
    sem_wait(draw_information->draw);

    if (!(draw_information->stop))
      update_screen(draw_information->bitmap);
    else
      stopped=1;

    sem_signal(draw_information->finished);
  }
}

int32 get_pvr_size(int32 size)
{
  if (size<=256)
    return 256;
  if (size<=512)
    return 512;
  return 1024;
}

uint32 cur_dsp_line  = 0;
int32  scr_width;

/*
Create a display screen, or window, of the given dimensions (or larger).
Attributes are the ones defined in driver.h.
Returns 0 on success.
*/
int osd_create_display(const struct osd_create_params *params, UINT32 *rgb_components)
{
  int pvr_height = get_pvr_size(params->height);
  int pvr_width  = get_pvr_size(params->width);

  scr_width = pvr_width*2;

	video_depth = params->depth==15?16:params->depth;

  printf(">> osd_create_display\n");
  printf(" -- osd_create_params:\n");
  printf(" -- width=%d, height=%d\n",params->width,params->height);
  printf(" -- aspect_x=%d, aspect_y=%d\n",params->aspect_x,params->aspect_y);
  printf(" -- depth=%d\n",params->depth);
  printf(" -- colors=%d\n",params->colors);
  printf(" -- fps=%f\n",(double)params->fps);
  printf(" -- video_attributes=0x%08X\n",params->video_attributes);
  printf(" -- orientation=%d\n",params->orientation);

	vsync_frame_rate = video_fps = (int)(params->fps);
	video_attributes = params->video_attributes;
	video_orientation = params->orientation;

  //total_ticks_last_cycle = 0;
  //ticks_per_cycle = (int)((double)TICKS_PER_SEC/params->fps);

	brightness               = 100;
	brightness_paused_adjust = 1.0;

  set_brightness((int)brightness);

	if (frameskip < 0) 
    frameskip = 0;
	if (frameskip >= FRAMESKIP_LEVELS) 
    frameskip = FRAMESKIP_LEVELS-1;
	
  frames_displayed = 0;
  start_time = 0;
  
  /* initialise the lookup tables */

  memset(&simple_bitmap,0x00,481*sizeof(int32));
  memset(&simple_display,0x00,480*sizeof(int32));
  memset(&simple_display2,0x00,480*sizeof(int32));

  /* ------------------------------------------------------ */
  /* Now allocate a pvr buffer suitable for the bitmap size */
  /* ------------------------------------------------------ */

  dc_texture_addr = pvr_mem_malloc(pvr_width*pvr_height*2);

  printf(" -- dc_texture_addr=%p\n -- pvr_width=%d pvr_height=%d\n",dc_texture_addr,(int)pvr_width,(int)pvr_height);

  if (dc_texture_addr)
  {
    float u1 = 0.0f;
    float v1 = 0.0f;
    float u2 = (float)params->width/(float)pvr_width;
    float v2 = (float)params->height/(float)pvr_height;

    int dsp_width  = params->width;
    int dsp_height = params->height;

    cur_dsp_line = ((uint32)dc_texture_addr) & ~0xA0000000;

    /* -------------------------------------------- */
    /* Switch width and height for display purposes */
    /* -------------------------------------------- */

    if (screen_orientation & 0x1)
    {
      dsp_width  = params->height;
      dsp_height = params->width;
    }

    /* ------------------------------- */
    /* choose the scene drawing points */
    /* ------------------------------- */

    switch(screen_orientation)
    {
      case 1:                /* 90 degrees */
        dc_u1 = dc_u2 = u1;
        dc_v2 = dc_v4 = v1;
        dc_u3 = dc_u4 = u2;
        dc_v1 = dc_v3 = v2;
        break;
      case 2:                /* 180 degrees */
        dc_u2 = dc_u4 = u1;
        dc_v3 = dc_v4 = v1;
        dc_u1 = dc_u3 = u2;
        dc_v1 = dc_v2 = v2;
        break;
      case 3:                /* 270 degrees */
        dc_u3 = dc_u4 = u1;
        dc_v1 = dc_v3 = v1;
        dc_u1 = dc_u2 = u2;
        dc_v2 = dc_v4 = v2;
        break;
      default:               /* 0 degrees */
        dc_u1 = dc_u3 = u1;
        dc_v1 = dc_v2 = v1;
        dc_u2 = dc_u4 = u2;
        dc_v3 = dc_v4 = v2;
        break;
    }

    pvr_poly_cxt_t cxt;

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_NONTWIDDLED | PVR_TXRFMT_RGB565,
                     pvr_width, pvr_height, dc_texture_addr, 
                     use_display_mode?PVR_FILTER_NONE:PVR_FILTER_BILINEAR);

    /* --------------------------------------------- */
    /* Display mode 3 == 0 with no graphic filtering */
    /* --------------------------------------------- */

    if (use_display_mode == 3)
      use_display_mode = 0;

    cxt.gen.shading = PVR_SHADE_FLAT;
    cxt.depth.write = PVR_DEPTHWRITE_DISABLE;
    
    pvr_poly_compile (&dc_pvr_poly_header, &cxt);

    /* --------------------------------------------------------------------------- */
    /* Calculate u and v values based on visible area                              */
    /* --------------------------------------------------------------------------- */

    printf(" -- u1=%f v1=%f\n",(double)u1,(double)v1);
    printf(" -- u2=%f v2=%f\n",(double)u2,(double)v2);

    /* ---------------------------------------------------------------------- */
    /* Calculate x and y values in order to scale and orient screen correctly */
    /* ---------------------------------------------------------------------- */

    if (use_display_mode == 0)     /* Stretch mode */
    {
      #define WT_MIN(a,b)((a)<(b)?(a):(b))
      #define WT_MAX(a,b)((a)>(b)?(a):(b))

      float aspect_ratio = (float)WT_MIN(dsp_width,dsp_height)/(float)WT_MAX(dsp_width,dsp_height);

      if (dsp_width > dsp_height)
      {
        float wid_diff = 639.0f - (float)dsp_width;

        dc_x2 = 639;
        dc_y2 = dsp_height+(int)(wid_diff*aspect_ratio);
      }
      else
      {
        float hgt_diff = 479.0f - (float)dsp_height;

        dc_y2 = 479;
        dc_x2 = dsp_width+(int)(hgt_diff*aspect_ratio);
      }

      if (dc_x2>639)
      {
        float wid_diff = 639.0f - (float)dc_x2;

        dc_x2 = 639;
        dc_y2 = 479+(int)(wid_diff*aspect_ratio);
      }
      else if (dc_y2>479)
      {
        float hgt_diff = 479.0f - (float)dc_y2;

        dc_y2 = 479;
        dc_x2 = 639+(int)(hgt_diff*aspect_ratio);
      }

      dc_x1 = (639-dc_x2)/2;
      dc_x2 += dc_x1;
      dc_y1 = (479-dc_y2)/2;
      dc_y2 += dc_y1;
    }
    else                           /* Nx mode */
    {
      int32 dsp_wid = dsp_width*use_display_mode;
      int32 dsp_hgt = dsp_height*use_display_mode;

      printf(" -- vis_wid: %d vis_hgt: %d use_display_mode: %d\n",(int)dsp_wid,(int)dsp_hgt,use_display_mode);

      dc_x1 = (639-dsp_wid)/2;
      dc_y1 = (479-dsp_hgt)/2;

      dc_x2 = dc_x1+dsp_wid;
      dc_y2 = dc_y1+dsp_hgt;
    }
    
    printf("onscreen1: %d,%d -> %d,%d (%d/%d)\n",dc_x1,dc_y1,dc_x2,dc_y2,(dc_x2+1-dc_x1),(dc_y2+1-dc_y1));
  }

  /* -------------------------------------------------------------------------- */
  /* The amount we want to copy is in fact width 16-bit values (assumes sq_cpy) */
  /* -------------------------------------------------------------------------- */
  
  simple_size  = pvr_width * params->height * 2;      //Only used in DMA mode
  simple_width = (((params->width*2)+0x1f)>>5)+1;

  /* ------------------------------------ */
  /* Tell MAME what it can use for the UI */
  /* ------------------------------------ */

  set_ui_visarea(0,0,params->width,params->height);

  /* ------------------------------------- */
  /* Launch the rendering thread if needed */
  /* ------------------------------------- */

  if (use_video_thread)
  {
    if ((draw_information = malloc(sizeof(struct DRAW_INFORMATION))))
    {
      memset(draw_information,0x00,sizeof(struct DRAW_INFORMATION));

      if ((draw_information->draw = sem_create(0)) &&
          (draw_information->finished = sem_create(0)))
      {
        if ((draw_information->draw_thread = thd_create(thd_dc_draw_screen, NULL)))
        {
          printf("VT%d\n",draw_information->draw_thread->tid);
        }
        else
        {
          sem_destroy(draw_information->draw);
          sem_destroy(draw_information->finished);
          free(draw_information);
          draw_information=0;
          use_video_thread = 0;
        }
      }
      else
      {
        free(draw_information);
        draw_information=0;
        use_video_thread = 0;
      }
    }
    else 
      use_video_thread = 0;
  }

  /* --------------------------- */
  /* Choose a rendering function */
  /* --------------------------- */

  if (use_dma)
    update_screen = update_screen_dma;
  else
    update_screen = update_screen_non_dma;

  /* ----------- */
  /* Colours ??? */
  /* ----------- */

  if (rgb_components)
  {
    rgb_components[0] = 0xF800;
    rgb_components[1] = 0x07E0;
    rgb_components[2] = 0x001F;
  }

  #if defined(WT_VBLANK_POLL_JOYSTICK)
  pj_handler = vblank_handler_add(poll_joysticks);
  #endif

  return 0;
}

/* osd_close_display =============================================== */
/*                                                                   */
/* =============================================== osd_close_display */

void osd_close_display(void)
{
  /* ------------------------------- */
  /* Tell the drawing thread to stop */
  /* ------------------------------- */

  if (use_video_thread)
  {
    draw_information->stop = 1;
    sem_signal(draw_information->draw);
    sem_wait(draw_information->finished);
    thd_destroy(draw_information->draw_thread);
    sem_destroy(draw_information->draw);
    sem_destroy(draw_information->finished);
  }

  /* ---------------------- */
  /* Now free everything up */
  /* ---------------------- */

	free(pens_lookup);
	current_pens_lookup = pens_lookup = 0;

	//free(debug_pens_lookup);
	//debug_pens_lookup = 0;

	//free(dirtycolor);
	//dirtycolor = 0;

	//free(current_palette);
	//current_palette = 0;

  free(draw_information);
  draw_information = 0;

  pvr_mem_free(dc_texture_addr);
  dc_texture_addr = 0;

  if (start_time && frames_displayed)
  {
    double fake_frames_displayed = (double)frames_displayed + (double)(frames_displayed/FRAMESKIP_LEVELS*frameskip);

    avg_fps = fake_frames_displayed/((double)(ticker()-start_time)/(double)TICKS_PER_SEC);

    printf("Average FPS: %f\nCurrent Frameskip: %d (auto:%d)\nframes_displayed: %d (ff:%f sec:%f)\n",
           (double)avg_fps,
           frameskip,autoframeskip,
           frames_displayed,fake_frames_displayed,((double)(ticker()-start_time)/(double)TICKS_PER_SEC));
  }
 
  #if defined(WT_VBLANK_POLL_JOYSTICK)
  vblank_handler_remove(pj_handler);
  pj_handler = 0;
  #endif
  
  return;
}

static void update_palette(struct mame_display *display)
{
  if (!pens_lookup)
    pens_lookup = malloc(display->game_palette_entries*sizeof(UINT16));

  if (pens_lookup)
  {
    int i, j;

    //printf("PALETTE ENTRIES: %d\n",display->game_palette_entries);

    // loop over dirty colors in batches of 32
	  for (i = 0; i < display->game_palette_entries; i += 32)
	  {
		  UINT32 dirtyflags = display->game_palette_dirty[i / 32];

		  if (dirtyflags)
		  {
			  display->game_palette_dirty[i / 32] = 0;

			  // loop over all 32 bits and update dirty entries
			  for (j = 0; (i+j < display->game_palette_entries) && (j < 32); j++, dirtyflags >>= 1)
				  if (dirtyflags & 1)
				  {
					  // extract the RGB values
					  rgb_t rgbvalue = display->game_palette[i + j];
					  int r = RGB_RED(rgbvalue);
					  int g = RGB_GREEN(rgbvalue);
					  int b = RGB_BLUE(rgbvalue);

					  // update lookup table
            pens_lookup[i+j] =  ((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0);
          }
			}
		}

    current_pens_lookup = pens_lookup;
	}
}

int osd_skip_this_frame(void)
{
	static const int skiptable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
	{
		{ 0,0,0,0,0,0,0,0,0,0,0,0 },
		{ 0,0,0,0,0,0,0,0,0,0,0,1 },
		{ 0,0,0,0,0,1,0,0,0,0,0,1 },
		{ 0,0,0,1,0,0,0,1,0,0,0,1 },
		{ 0,0,1,0,0,1,0,0,1,0,0,1 },
		{ 0,1,0,0,1,0,1,0,0,1,0,1 },
		{ 0,1,0,1,0,1,0,1,0,1,0,1 },
		{ 0,1,0,1,1,0,1,0,1,1,0,1 },
		{ 0,1,1,0,1,1,0,1,1,0,1,1 },
		{ 0,1,1,1,0,1,1,1,0,1,1,1 },
		{ 0,1,1,1,1,1,0,1,1,1,1,1 },
		{ 0,1,1,1,1,1,1,1,1,1,1,1 }
	};

	return skiptable[frameskip][frameskip_counter];
}

void osd_update_video_and_audio(struct mame_display *display)
{
	int i;
	//struct mame_bitmap *bitmap = display->game_bitmap;

  //TICKER total_ticks_this_cycle;
  //TICKER ticks_this_cycle;

  /* ------------------------------------------------------------------ */
  /* This is the first time through, fill in the (sneaky) lookup tables */
  /* ------------------------------------------------------------------ */

  if (!start_time)
  {
    /* ------------------------------------------------------------------------------ */
    /* Now fill our lookup tables with the correct mapping locations (assumes sq_cpy) */
    /* ------------------------------------------------------------------------------ */

    if (use_dma)
      simple_bitmap[0] = (uint32)display->game_bitmap->line[display->game_visible_area.min_y];
    else
    {
      for(i=display->game_visible_area.min_y;i<=display->game_visible_area.max_y;i++)
      {
        simple_bitmap[i-display->game_visible_area.min_y]   = (uint32)display->game_bitmap->line[i]+(display->game_visible_area.min_x*2);
        simple_display2[i-display->game_visible_area.min_y] = (uint32)(void *)(0xe0000000 | (((unsigned long)cur_dsp_line) & 0x03ffffe0));
        simple_display[i-display->game_visible_area.min_y]  = ((cur_dsp_line>>26)<<2)&0x1c;
        cur_dsp_line += scr_width;
      }

      /* --------------------------------------------------- */
      /* Put a sentinel value at the end of the bitmap table */
      /* --------------------------------------------------- */

      simple_bitmap[i-display->game_visible_area.min_y] = 0; 
    }

    /* ------------------------------------------------------------------------------------------ */
    /* Temporarily set a non-zero start_time, it actually gets set for real in osd_reset_counters */
    /* ------------------------------------------------------------------------------------------ */

    start_time = 1;
  }

  /* ------------------------------------------ */
  /* If the game palette has changed, update it */
  /* ------------------------------------------ */

	if (display->changed_flags & GAME_PALETTE_CHANGED)
		update_palette(display);

  /* ------------------------------------------ */
  /* If we're not skipping this frame... render */
  /* ------------------------------------------ */

	if (display->changed_flags & GAME_BITMAP_CHANGED)
	{
		//if (input_ui_pressed(IPT_UI_SHOW_FPS))
    //  showfps = !showfps;

		profiler_mark(PROFILER_BLIT);
    
    if (use_video_thread)
    {
      draw_information->bitmap = display->game_bitmap;

      sem_signal(draw_information->draw);

      //if (use_throttle)
      //  sem_wait(draw_information->finished);
    }
    else
		  update_screen(display->game_bitmap);

    if (update_audio)
      update_audio(play_buffer);

		profiler_mark(PROFILER_END);

    /* ----------------- */
    /* Get some stats... */
    /* ----------------- */

    //total_ticks_this_cycle = ticker();

    //ticks_this_cycle = total_ticks_this_cycle-total_ticks_last_cycle;

    ++frames_displayed;
    
    //if (!frameskip_counter)
    //  speed = (int)(((double)ticks_per_cycle/(double)ticks_this_cycle)*100.0);

    //if (use_throttle && !frameskip && ticks_this_cycle<ticks_per_cycle)
    //{
    //  int delay = (int)(((double)(ticks_per_cycle-ticks_this_cycle)/TICKS_PER_MS)+0.5);
    //
    //  profiler_mark(PROFILER_IDLE);
    //
    //  if (delay>2) timer_spin_sleep(--delay);
    //
    //  profiler_mark(PROFILER_END);
    //
    //  /* ---------------------------------------------------------- */
    //  /* Capture the tick count now, don't want to count the sleep! */
    //  /* ---------------------------------------------------------- */
    //
    //  total_ticks_last_cycle = ticker();
    //}
    //else
    //{
    //  /* ------------------------------------------------- */
    //  /* It's close enough!!! and saves a call to ticker() */
    //  /* ------------------------------------------------- */
    //
    //  total_ticks_last_cycle = total_ticks_this_cycle;
    //}
	}
  else
  { 
    /* ------------------------------------------ */
    /* If we're playing audio now... play it now! */
    /* ------------------------------------------ */

    if (update_audio)
      update_audio(play_buffer);
  }
  
  /* ------------------------- */
  /* Do any autoframeskip work */
  /* ------------------------- */

  #if 0                            /* AUTO not working */
  if (autoframeskip && !frameskip_counter)
  {
    static int frameskipadjust;
    int adjspeed;

    /* adjust speed to video refresh rate if vsync is on */
    adjspeed = speed * video_fps / vsync_frame_rate;

    if (adjspeed >= 105)
    {
      frameskipadjust++;
      if (frameskipadjust >= 3)
      {
        frameskipadjust = 0;
        if (frameskip > 0) frameskip--;
      }
    }
    else
    {
      if (adjspeed < 80)
        frameskipadjust -= (90 - adjspeed) / 5;
      else
      {
        /* don't push frameskip too far if we are close to 100% speed */
        if (frameskip < 8)
          frameskipadjust--;
      }

      while (frameskipadjust <= -2)
      {
        frameskipadjust += 2;
        /* Digita, we don't work so well with frameskip of 12 */
        if (frameskip < 8) frameskip++;
      }
    }
  }
  #endif

  /* --------------------- */
  /* UI Framskip Increment */
  /* --------------------- */

  #if 0 //OVERCLOCK_STUFF
  if (input_ui_pressed(IPT_UI_PAN_UP))
    timer_set_overclock(0,timer_get_overclock(0)+0.01);
  else if (input_ui_pressed(IPT_UI_PAN_DOWN))
    timer_set_overclock(0,timer_get_overclock(0)-0.01);

  if (input_ui_pressed(IPT_UI_PAN_LEFT))
    timer_set_overclock(1,timer_get_overclock(1)+0.01);
  else if (input_ui_pressed(IPT_UI_PAN_RIGHT))
    timer_set_overclock(1,timer_get_overclock(1)-0.01);
  #endif
  
  #if 0
	if (input_ui_pressed(IPT_UI_FRAMESKIP_INC))
	{
		if (autoframeskip)
		{
			autoframeskip = 0;
			frameskip = 0;
		}
		else
		{
			if (frameskip == FRAMESKIP_LEVELS-1)
			{
				frameskip = 0;
				autoframeskip = 1;
			}
			else
				frameskip++;
		}

		frames_displayed = 0;
	}

  /* --------------------- */
  /* UI Framskip Decrement */
  /* --------------------- */

	if (input_ui_pressed(IPT_UI_FRAMESKIP_DEC))
	{
		if (autoframeskip)
		{
			autoframeskip = 0;
			frameskip = FRAMESKIP_LEVELS-1;
		}
		else
		{
			if (frameskip == 0)
				autoframeskip = 1;
			else
				frameskip--;
		}

		frames_displayed = 0;
	}

  /* ----------- */
  /* UI Throttle */
  /* ----------- */

	if (input_ui_pressed(IPT_UI_THROTTLE))
	{
		use_throttle = !use_throttle;

		frames_displayed = 0;
	}
  #endif
  
	frameskip_counter = (frameskip_counter + 1) % FRAMESKIP_LEVELS;

  #if !defined(WT_VBLANK_POLL_JOYSTICK)
  poll_joysticks(0);
  #endif

  WT_PEXIT(2);   /* Yes - these are the wrong way around - it's the entire loop! */
  WT_PENTRY(2);

  WT_PDISPLAY;
}

/* osd_get_fps_text ================================================ */
/*                                                                   */
/* ================================================ osd_get_fps_text */

const char *osd_get_fps_text(const struct performance_info *performance)
{
  static char buf[30];

  sprintf(buf,"0:%d 1:%d %s%2d%f%%",(int)(timer_get_overclock(0)*100.0f),(int)(timer_get_overclock(1)*100.0f),autoframeskip?"auto":"fskp",frameskip,performance->game_speed_percent);
  
  return buf;
}

/* osd_cycles ====================================================== */
/*                                                                   */
/* ====================================================== osd_cycles */

cycles_t osd_cycles(void)
{
  return uclock();
}

/* osd_cycles_per_second =========================================== */
/*                                                                   */
/* =========================================== osd_cycles_per_second */

cycles_t osd_cycles_per_second(void)
{
  return UCLOCKS_PER_SEC;
}

/* osd_profiling_ticks ============================================= */
/*                                                                   */
/* ============================================= osd_profiling_ticks */

cycles_t osd_profiling_ticks(void)
{
  return ticker();
}

int raise(int r)
{
  printf("raise %d\n",r);
}

/* set_brightness ============================================== */
/*                                                               */
/* brightess = percentage 0-100%                                 */
/*                                                               */
/* ============================================== set_brightness */

void set_brightness(int _brightness)
{
	//brightness = (float)_brightness;
  
  dc_brightness = (_brightness/100.0f)*brightness_paused_adjust;

  return;
}

/* get_brightness ============================================== */
/*                                                               */
/* ============================================== get_brightness */

int get_brightness(void)
{
	return (int)brightness;
}

/* osd_save_snapshot =============================================== */
/*                                                                   */
/* =============================================== osd_save_snapshot */

void osd_save_snapshot(struct mame_bitmap *bitmap, const struct rectangle *bounds)
{
  vid_screen_shot("/pc/tmp/mdc_screenshot.ppm");

  return;
}

/* osd_pause ======================================================= */
/*                                                                   */
/* ======================================================= osd_pause */

void osd_pause(int paused)
{
	if (paused)
    brightness_paused_adjust=0.65f;
  else
    brightness_paused_adjust=1.0f;

  return(set_brightness((int)brightness));
}

/* osd_mark_vector_dirty =========================================== */
/*                                                                   */
/* =========================================== osd_mark_vector_dirty */

void osd_mark_vector_dirty(int x, int y)
{
  return;
}   

